!#######################################################
!test program to send derived data types through MPI_Send / MPI_Recv
!The program uses some arbitrary data structures.
!
!To add / delete components in the data structures change in 
!the corresponding definition, allocation/initailization, pack 
!and unpack routines 
!
!compile:
!module load intel/12.1.4  impi/4.0.3.008
!mpif90 -fpp pack_unpack.f90
!run on 2 or more ranks, it will use only 0 and 1 rank for send /recv
!
!Author : Chandan Basu, cbasu@nsc.liu.se
!#######################################################



module derived_type
use mpi

implicit none
integer :: bytes, pos
integer :: MPI_DERIVED_DATA

private :: bytes, pos
public :: MPI_DERIVED_DATA

type l 
        ! any change here will require changes in allocation,
        ! initalization,
        ! pack_type_l and unpack_type_l
        integer :: n1, n2, n3
        double precision :: d1, d2, d3
end type l
public :: l

type band
        ! any change here will require changes in allocation,
        ! initalization,
        ! pack_type_band and unpack_type_band
        integer :: n1, n2, n3
        ! nl is the no. of l. Storing this info in data type is very
        ! handy.
        integer :: nl                         
        type(l), dimension(:), allocatable :: l
end type band
public :: band

type point
        ! any change here will require changes in allocation,
        ! initalization,
        ! pack_type_point and unpack_type_point
        integer :: n1, n2, n3
        ! nband is the no. of band. Storing this info in data type is
        ! very handy.
        integer :: nband                      
        type(band), dimension(:), allocatable :: band
end type point
public :: point

public :: MPI_Send_point, MPI_recv_point

private :: pack_type_point, unpack_type_point 
private :: pack_type_band, unpack_type_band
private :: pack_type_l, unpack_type_l

contains

subroutine MPI_Send_point(pt, npt, dummy, tgt, tag, comm, ierr)
type(point), dimension(:) :: pt
integer :: npt, dummy, tgt, tag, comm, ierr
integer :: i, j, k
character, dimension(:), allocatable :: buffer

bytes = 0
pos = 0
do i = 1, npt
        bytes = bytes + 4 * sizeof (pt(i)%n1)
        do j = 1, pt(i)%nband
                bytes = bytes + 4 * sizeof (pt(i)%band(j)%n1)
                do k = 1, pt(i)%band(j)%nl
                        bytes = bytes + 3 * sizeof
(pt(i)%band(j)%l(k)%n1)
                        bytes = bytes + 3 * sizeof
(pt(i)%band(j)%l(k)%d1)
                end do
        end do
end do
allocate(buffer(bytes))       !allocate buffer of same size as the data
type point
call pack_type_point(buffer, pt, npt, comm) !  pack pt in buffer 
call MPI_Send(buffer, bytes, MPI_CHARACTER, tgt, tag, comm, ierr)
deallocate(buffer)

end subroutine MPI_Send_point

subroutine MPI_recv_point(pt, npt, dummy, src, tag, comm, stat, ierr)
type(point), dimension(:) :: pt
integer :: npt, dummy, src, tag, comm, stat, ierr
integer :: i, j, k
character, dimension(:), allocatable :: buffer

bytes = 0
pos = 0

do i = 1, npt
        bytes = bytes + 4 * sizeof (pt(i)%n1)
        do j = 1, pt(i)%nband
                bytes = bytes + 4 * sizeof (pt(i)%band(j)%n1)
                do k = 1, pt(i)%band(j)%nl
                        bytes = bytes + 3 * sizeof
(pt(i)%band(j)%l(k)%n1)
                        bytes = bytes + 3 * sizeof
(pt(i)%band(j)%l(k)%d1)
                end do
        end do
end do

allocate(buffer(bytes))       !allocate buffer of same size as the data
type point
call MPI_Recv(buffer, bytes, MPI_CHARACTER, src, tag, comm, stat, ierr)
call unpack_type_point(buffer, pt, npt, comm)  ! unpack
deallocate(buffer)
end subroutine MPI_Recv_point

subroutine pack_type_point(buf, v, nv, comm)
character, dimension (:), intent(inout) :: buf
type(point), dimension(:) :: v
integer :: nv, comm
integer :: i, j, ierr
        do i = 1, nv 
                call MPI_Pack(int(v(i)%n1), 1, MPI_INTEGER, buf, bytes,
pos, MPI_COMM_WORLD, ierr);
                call MPI_Pack(int(v(i)%n2), 1, MPI_INTEGER, buf, bytes,
pos, MPI_COMM_WORLD, ierr);
                call MPI_Pack(int(v(i)%n3), 1, MPI_INTEGER, buf, bytes,
pos, MPI_COMM_WORLD, ierr);
                call MPI_Pack(int(v(i)%nband), 1, MPI_INTEGER, buf,
bytes, pos, MPI_COMM_WORLD, ierr);
                do j = 1, v(i) % nband
                        call pack_type_band(buf, pos, v(i)%band(j),
MPI_COMM_WORLD)
                end do
        end do
end subroutine pack_type_point

subroutine unpack_type_point(buf, v, nv, comm)
character, dimension (:):: buf
type(point), dimension(:), intent(inout) :: v
integer :: nv, comm
integer :: i, j, ierr
        do i = 1, nv 
                call MPI_Unpack(buf, bytes, pos, v(i)%n1, 1,
MPI_INTEGER, MPI_COMM_WORLD, ierr)
                call MPI_Unpack(buf, bytes, pos, v(i)%n2, 1,
MPI_INTEGER, MPI_COMM_WORLD, ierr)
                call MPI_Unpack(buf, bytes, pos, v(i)%n3, 1,
MPI_INTEGER, MPI_COMM_WORLD, ierr)
                call MPI_Unpack(buf, bytes, pos, v(i)%nband, 1,
MPI_INTEGER, MPI_COMM_WORLD, ierr)
                do j = 1, v(i)%nband
                        call unpack_type_band(buf, pos, v(i)%band(j),
MPI_COMM_WORLD)
                end do
        end do
end subroutine unpack_type_point

subroutine pack_type_band(buf, pos, v, comm)
character, dimension (:), intent(inout) :: buf
integer :: pos
type(band) :: v
integer :: comm
integer :: i, ierr

        call MPI_Pack(int(v%n1), 1, MPI_INTEGER, buf, bytes, pos,
MPI_COMM_WORLD, ierr);
        call MPI_Pack(int(v%n2), 1, MPI_INTEGER, buf, bytes, pos,
MPI_COMM_WORLD, ierr);
        call MPI_Pack(int(v%n3), 1, MPI_INTEGER, buf, bytes, pos,
MPI_COMM_WORLD, ierr);
        call MPI_Pack(int(v%nl), 1, MPI_INTEGER, buf, bytes, pos,
MPI_COMM_WORLD, ierr);
        do i = 1, v%nl
                call pack_type_l(buf, pos, v%l(i), MPI_COMM_WORLD)
        end do
end subroutine pack_type_band

subroutine unpack_type_band(buf, pos, v, comm)
character, dimension (:):: buf
integer, intent(inout) :: pos
type(band), intent(inout) :: v
integer :: comm
integer :: i, ierr

        call MPI_Unpack(buf, bytes, pos, v%n1, 1, MPI_INTEGER,
MPI_COMM_WORLD, ierr)
        call MPI_Unpack(buf, bytes, pos, v%n2, 1, MPI_INTEGER,
MPI_COMM_WORLD, ierr)
        call MPI_Unpack(buf, bytes, pos, v%n3, 1, MPI_INTEGER,
MPI_COMM_WORLD, ierr)
        call MPI_Unpack(buf, bytes, pos, v%nl, 1, MPI_INTEGER,
MPI_COMM_WORLD, ierr)
        do i = 1, v%nl
                call unpack_type_l(buf, pos, v%l(i), MPI_COMM_WORLD)
        end do
end subroutine unpack_type_band

subroutine pack_type_l(buf, pos, v, comm)
character, dimension (:), intent(inout) :: buf
integer :: pos
type(l) :: v
integer :: comm, ierr

        call MPI_Pack(int(v%n1), 1, MPI_INTEGER, buf, bytes, pos,
MPI_COMM_WORLD, ierr);
        call MPI_Pack(int(v%n2), 1, MPI_INTEGER, buf, bytes, pos,
MPI_COMM_WORLD, ierr);
        call MPI_Pack(int(v%n3), 1, MPI_INTEGER, buf, bytes, pos,
MPI_COMM_WORLD, ierr);
        call MPI_Pack(v%d1, 1, MPI_DOUBLE_PRECISION, buf, bytes, pos,
MPI_COMM_WORLD, ierr);
        call MPI_Pack(v%d2, 1, MPI_DOUBLE_PRECISION, buf, bytes, pos,
MPI_COMM_WORLD, ierr);
        call MPI_Pack(v%d3, 1, MPI_DOUBLE_PRECISION, buf, bytes, pos,
MPI_COMM_WORLD, ierr);
end subroutine pack_type_l

subroutine unpack_type_l(buf, pos, v, comm)
character, dimension (:), intent(in) :: buf
integer, intent(inout) :: pos
type(l), intent(inout) :: v
integer :: comm, ierr
        call MPI_Unpack(buf, bytes, pos, v%n1, 1, MPI_INTEGER,
MPI_COMM_WORLD, ierr)
        call MPI_Unpack(buf, bytes, pos, v%n2, 1, MPI_INTEGER,
MPI_COMM_WORLD, ierr)
        call MPI_Unpack(buf, bytes, pos, v%n3, 1, MPI_INTEGER,
MPI_COMM_WORLD, ierr)
        call MPI_Unpack(buf, bytes, pos, v%d1, 1, MPI_DOUBLE_PRECISION,
MPI_COMM_WORLD, ierr)
        call MPI_Unpack(buf, bytes, pos, v%d2, 1, MPI_DOUBLE_PRECISION,
MPI_COMM_WORLD, ierr)
        call MPI_Unpack(buf, bytes, pos, v%d3, 1, MPI_DOUBLE_PRECISION,
MPI_COMM_WORLD, ierr)
        end subroutine unpack_type_l
end module derived_type

program main
use ifport, only : irand, rand
use derived_type
implicit none
integer, parameter :: NPOINT = 100       ! no. of point data types

integer :: i,j,k

type (point), dimension(:), allocatable :: var1, var2
integer :: ierr, rank, stat, nerr

call MPI_Init(ierr)
call MPI_Comm_rank(MPI_COMM_WORLD, rank, ierr)

! allocate both var1 & var2, initialize var1 with random data
allocate (var1(NPOINT))  
allocate (var2(NPOINT))  

do i = 1, NPOINT
        var1(i)%n1 = irand()
        var1(i)%n2 = irand()
        var1(i)%n3 = irand()
        var1(i)%nband = i+10          ! some value for number of bands
        var2(i)%nband = i+10          ! some value for number of bands
        allocate(var1(i) % band(var1(i)%nband))
        allocate(var2(i) % band(var2(i)%nband))
        do j = 1, var1(i)%nband
                var1(i)%band(j)%n1 = irand()
                var1(i)%band(j)%n2 = irand()
                var1(i)%band(j)%n3 = irand()
                var1(i)%band(j)%nl = j + 10  ! some value for nl
                var2(i)%band(j)%nl = j + 10  ! some value for nl
                allocate(var1(i) % band(j) % l(var1(i)%band(j)%nl))
                allocate(var2(i) % band(j) % l(var2(i)%band(j)%nl))
                do k = 1, var1(i)%band(j)%nl
                        var1(i)%band(j)%l(k)%n1 = irand()
                        var1(i)%band(j)%l(k)%n2 = irand()
                        var1(i)%band(j)%l(k)%n3 = irand()
                        var1(i)%band(j)%l(k)%d1 = rand()
                        var1(i)%band(j)%l(k)%d2 = rand()
                        var1(i)%band(j)%l(k)%d3 = rand()
                end do
        end do
end do

if (rank == 0) then
        call MPI_Send_point(var1, NPOINT, MPI_DERIVED_DATA, 1, 100,
MPI_COMM_WORLD, ierr)
else if (rank == 1) then
        call MPI_Recv_point(var2, NPOINT, MPI_DERIVED_DATA, 0, 100,
MPI_COMM_WORLD, stat, ierr)

        ! do a test if the values are correctly sent
        nerr = 0
        do i = 1, NPOINT
                if (var2(i)%n1 /= var1(i)%n1) then
                        nerr = nerr + 1
                else if (var2(i)%n2 /= var1(i)%n2) then
                        nerr = nerr + 1
                else if (var2(i)%n3 /= var1(i)%n3) then
                        nerr = nerr + 1
                else if (var2(i)%nband /= var1(i)%nband) then
                        nerr = nerr + 1
                end if
                do j = 1, var1(i) % nband
                        if (var2(i)%band(j)%n1 /= var1(i)%band(j)%n1)
then
                                nerr = nerr + 1
                        else if (var2(i)%band(j)%n2 /=
var1(i)%band(j)%n2) then
                                nerr = nerr + 1
                        else if (var2(i)%band(j)%n3 /=
var1(i)%band(j)%n3) then
                                nerr = nerr + 1
                        else if (var2(i)%band(j)%nl /=
var1(i)%band(j)%nl) then
                                nerr = nerr + 1
                        end if
                        do k = 1, var2(i)%band(j)%nl
                                if (var2(i)%band(j)%l(k)%n1 /=
var1(i)%band(j)%l(k)%n1) then
                                        nerr = nerr + 1
                                else if (var2(i)%band(j)%l(k)%n2 /=
var1(i)%band(j)%l(k)%n2) then
                                        nerr = nerr + 1
                                else if (var2(i)%band(j)%l(k)%n3 /=
var1(i)%band(j)%l(k)%n3) then
                                        nerr = nerr + 1
                                else if (var2(i)%band(j)%l(k)%d1 /=
var1(i)%band(j)%l(k)%d1) then
                                        nerr = nerr + 1
                                else if (var2(i)%band(j)%l(k)%d2 /=
var1(i)%band(j)%l(k)%d2) then
                                        nerr = nerr + 1
                                else if (var2(i)%band(j)%l(k)%d3 /=
var1(i)%band(j)%l(k)%d3) then
                                        nerr = nerr + 1
                                end if
                        end do
                end do
        end do
        if (nerr == 0) then
                print *, "Data copied correctly"
        else
                print *, "Error in data copy"
        end if
end if
do i = 1, NPOINT
        do j = 1, var1(i)%nband
                deallocate(var1(i) % band(j) % l)
                deallocate(var2(i) % band(j) % l)
        end do
        deallocate(var1(i) % band)
        deallocate(var2(i) % band)
end do
deallocate(var1)
deallocate(var2)
call MPI_Finalize(ierr)

end program main
